@@ -1,22 +1,24 @@
 import mistune
-import schema
+from . import schema
 import json
-import yaml
 import os
 import copy
 import re
 import sys
-import StringIO
+from StringIO import StringIO
 import logging
 import urlparse
-from aslist import aslist
+from .aslist import aslist
+from .add_dictlist import add_dictlist
 import re
 import argparse
+from typing import Any, IO, Union
 
 _logger = logging.getLogger("salad")
 
-def has_types(items):
-    r = []
+
+def has_types(items):  # type: (Any) -> List[basestring]
+    r = []  # type: List
     if isinstance(items, dict):
         if items["type"] == "https://w3id.org/cwl/salad#record":
             return [items["name"]]
@@ -32,36 +34,52 @@ def has_types(items):
         return [items]
     return []
 
+
 def linkto(item):
     _, frg = urlparse.urldefrag(item)
     return "[%s](#%s)" % (frg, to_id(frg))
 
-class MyRenderer(mistune.Renderer):
-    def header(self, text, level, raw=None):
-        return """<h%i id="%s">%s</h1>""" % (level, to_id(text), text)
 
-def to_id(text):
+class MyRenderer(mistune.Renderer):
+
+    def __init__(self):  # type: () -> None
+        super(mistune.Renderer, self).__init__()
+        self.options = {}
+
+    def header(self, text, level, raw=None):
+        return """<h%i id="%s">%s</h%i>""" % (level, to_id(text), text, level)
+
+    def table(self, header, body):
+        return (
+            '<table class="table table-striped">\n<thead>%s</thead>\n'
+            '<tbody>\n%s</tbody>\n</table>\n'
+        ) % (header, body)
+
+
+def to_id(text):  # type: (Union[str, unicode]) -> Union[str, unicode]
     textid = text
     if text[0] in ("0", "1", "2", "3", "4", "5", "6", "7", "8", "9"):
         try:
-            textid = text[text.index(" ")+1:]
+            textid = text[text.index(" ") + 1:]
         except ValueError:
             pass
     textid = textid.replace(" ", "_")
     return textid
 
+
 class ToC(object):
-    def __init__(self):
+
+    def __init__(self):  # type: () -> None
         self.first_toc_entry = True
         self.numbering = [0]
         self.toc = ""
         self.start_numbering = True
 
-    def add_entry(self, thisdepth, title):
+    def add_entry(self, thisdepth, title):  # type: (int, str) -> str
         depth = len(self.numbering)
         if thisdepth < depth:
             self.toc += "</ol>"
-            for n in range(0, depth-thisdepth):
+            for n in range(0, depth - thisdepth):
                 self.numbering.pop()
                 self.toc += "</li></ol>"
             self.numbering[-1] += 1
@@ -75,22 +93,24 @@ class ToC(object):
             self.numbering.append(1)
 
         if self.start_numbering:
-            num = "%i.%s" % (self.numbering[0], ".".join([str(n) for n in self.numbering[1:]]))
+            num = "%i.%s" % (self.numbering[0], ".".join(
+                [str(n) for n in self.numbering[1:]]))
         else:
             num = ""
-        self.toc += """<li><a href="#%s">%s %s</a><ol>\n""" %(to_id(title),
-            num, title)
+        self.toc += """<li><a href="#%s">%s %s</a><ol>\n""" % (to_id(title),
+                                                               num, title)
         return num
 
-    def contents(self, id):
+    def contents(self, idn):  # type: (str) -> str
         c = """<h1 id="%s">Table of contents</h1>
-               <nav class="tocnav"><ol>%s""" % (id, self.toc)
+               <nav class="tocnav"><ol>%s""" % (idn, self.toc)
         c += "</ol>"
         for i in range(0, len(self.numbering)):
             c += "</li></ol>"
         c += """</nav>"""
         return c
 
+
 basicTypes = ("https://w3id.org/cwl/salad#null",
               "http://www.w3.org/2001/XMLSchema#boolean",
               "http://www.w3.org/2001/XMLSchema#int",
@@ -102,12 +122,8 @@ basicTypes = ("https://w3id.org/cwl/salad#null",
               "https://w3id.org/cwl/salad#enum",
               "https://w3id.org/cwl/salad#array")
 
-def add_dictlist(di, key, val):
-    if key not in di:
-        di[key] = []
-    di[key].append(val)
 
-def number_headings(toc, maindoc):
+def number_headings(toc, maindoc):  # type: (ToC, str) -> str
     mdlines = []
     skip = False
     for line in maindoc.splitlines():
@@ -115,7 +131,7 @@ def number_headings(toc, maindoc):
             toc.start_numbering = True
             toc.numbering = [0]
 
-        if line == "```":
+        if "```" in line:
             skip = not skip
 
         if not skip:
@@ -129,27 +145,35 @@ def number_headings(toc, maindoc):
     maindoc = '\n'.join(mdlines)
     return maindoc
 
-def fix_doc(doc):
+
+def fix_doc(doc):  # type: (Union[List[str], str]) -> str
     if isinstance(doc, list):
-        doc = "".join(doc)
-    return "\n".join([re.sub(r"<([^>@]+@[^>]+)>", r"[\1](mailto:\1)", d) for d in doc.splitlines()])
+        docstr = "".join(doc)
+    else:
+        docstr = doc
+    return "\n".join(
+        [re.sub(r"<([^>@]+@[^>]+)>", r"[\1](mailto:\1)", d)
+         for d in docstr.splitlines()])
+
 
 class RenderType(object):
+
     def __init__(self, toc, j, renderlist, redirects):
-        self.typedoc = StringIO.StringIO()
+        # type: (ToC, List[Dict], str, Dict) -> None
+        self.typedoc = StringIO()
         self.toc = toc
-        self.subs = {}
-        self.docParent = {}
-        self.docAfter = {}
-        self.rendered = set()
+        self.subs = {}  # type: Dict[str, str]
+        self.docParent = {}  # type: Dict[str, List]
+        self.docAfter = {}  # type: Dict[str, List]
+        self.rendered = set()  # type: Set[str]
         self.redirects = redirects
-        self.title = None
+        self.title = None  # type: str
 
         for t in j:
             if "extends" in t:
                 for e in aslist(t["extends"]):
                     add_dictlist(self.subs, e, t["name"])
-                    #if "docParent" not in t and "docAfter" not in t:
+                    # if "docParent" not in t and "docAfter" not in t:
                     #    add_dictlist(self.docParent, e, t["name"])
 
             if t.get("docParent"):
@@ -162,12 +186,12 @@ class RenderType(object):
             if t.get("docAfter"):
                 add_dictlist(self.docAfter, t["docAfter"], t["name"])
 
-        _, _, metaschema_loader = schema.get_metaschema()
+        metaschema_loader = schema.get_metaschema()[2]
         alltypes = schema.extend_and_specialize(j, metaschema_loader)
 
-        self.typemap = {}
-        self.uses = {}
-        self.record_refs = {}
+        self.typemap = {}  # type: Dict
+        self.uses = {}  # type: Dict
+        self.record_refs = {}  # type: Dict
         for t in alltypes:
             self.typemap[t["name"]] = t
             try:
@@ -183,7 +207,7 @@ class RenderType(object):
                                 _, frg2 = urlparse.urldefrag(f["name"])
                                 self.uses[tp].append((frg1, frg2))
                             if tp not in basicTypes and tp not in self.record_refs[t["name"]]:
-                                    self.record_refs[t["name"]].append(tp)
+                                self.record_refs[t["name"]].append(tp)
             except KeyError as e:
                 _logger.error("Did not find 'type' in %s", t)
                 raise
@@ -196,16 +220,27 @@ class RenderType(object):
                  ("docAfter" not in f))):
                 self.render_type(f, 1)
 
-    def typefmt(self, tp, redirects, nbsp=False):
+    def typefmt(self, tp, redirects, nbsp=False, jsonldPredicate=None):
+        # type: (Any, Dict[str, str], bool, Dict[str, str]) -> Union[str, unicode]
         global primitiveType
         if isinstance(tp, list):
             if nbsp and len(tp) <= 3:
-                return "&nbsp;|&nbsp;".join([self.typefmt(n, redirects) for n in tp])
+                return "&nbsp;|&nbsp;".join([self.typefmt(n, redirects, jsonldPredicate=jsonldPredicate) for n in tp])
             else:
                 return " | ".join([self.typefmt(n, redirects) for n in tp])
         if isinstance(tp, dict):
             if tp["type"] == "https://w3id.org/cwl/salad#array":
-                return "array&lt;%s&gt;" % (self.typefmt(tp["items"], redirects, nbsp=True))
+                ar = "array&lt;%s&gt;" % (self.typefmt(tp["items"], redirects, nbsp=True))
+                if jsonldPredicate and "mapSubject" in jsonldPredicate:
+                    if "mapPredicate" in jsonldPredicate:
+                        ar += " | map&lt;%s.%s,&nbsp;%s.%s&gt" % (self.typefmt(tp["items"], redirects),
+                                                           jsonldPredicate["mapSubject"],
+                                                           self.typefmt(tp["items"], redirects),
+                                                           jsonldPredicate["mapPredicate"])
+                    ar += " | map&lt;%s.%s,&nbsp;%s&gt" % (self.typefmt(tp["items"], redirects),
+                                                          jsonldPredicate["mapSubject"],
+                                                          self.typefmt(tp["items"], redirects))
+                return ar
             if tp["type"] in ("https://w3id.org/cwl/salad#record", "https://w3id.org/cwl/salad#enum"):
                 frg = schema.avro_name(tp["name"])
                 if tp["name"] in redirects:
@@ -227,8 +262,7 @@ class RenderType(object):
                     tp = frg
                 return """<a href="#%s">%s</a>""" % (to_id(tp), tp)
 
-
-    def render_type(self, f, depth):
+    def render_type(self, f, depth):  # type: (Dict[str, Any], int) -> None
         if f["name"] in self.rendered or f["name"] in self.redirects:
             return
         self.rendered.add(f["name"])
@@ -244,6 +278,7 @@ class RenderType(object):
             f["doc"] = ""
 
         def extendsfrom(item, ex):
+            # type: (Dict[str, Any], List[Dict[str, Any]]) -> None
             if "extends" in item:
                 for e in aslist(item["extends"]):
                     ex.insert(0, self.typemap[e])
@@ -258,8 +293,9 @@ class RenderType(object):
                 for i in e["doc"]:
                     idx = i.find(":")
                     if idx > -1:
-                        enumDesc[i[:idx]] = i[idx+1:]
-                e["doc"] = [i for i in e["doc"] if i.find(":") == -1 or i.find(" ") < i.find(":")]
+                        enumDesc[i[:idx]] = i[idx + 1:]
+                e["doc"] = [i for i in e["doc"] if i.find(
+                    ":") == -1 or i.find(" ") < i.find(":")]
 
         f["doc"] = fix_doc(f["doc"])
 
@@ -282,21 +318,24 @@ class RenderType(object):
         else:
             doc = ""
 
-        if self.title is None:
-            self.title = f["doc"][0:f["doc"].index("\n")][2:]
+        if self.title is None and f["doc"]:
+            self.title = f["doc"][0:f["doc"].index("\n")]
+            if self.title.startswith('# '):
+                self.title = self.title[2:]
 
         if f["type"] == "documentation":
             f["doc"] = number_headings(self.toc, f["doc"])
 
-        #if "extends" in f:
+        # if "extends" in f:
         #    doc += "\n\nExtends "
         #    doc += ", ".join([" %s" % linkto(ex) for ex in aslist(f["extends"])])
-        #if f["name"] in self.subs:
+        # if f["name"] in self.subs:
         #    doc += "\n\nExtended by"
         #    doc += ", ".join([" %s" % linkto(s) for s in self.subs[f["name"]]])
-        #if f["name"] in self.uses:
+        # if f["name"] in self.uses:
         #    doc += "\n\nReferenced by"
-        #    doc += ", ".join([" [%s.%s](#%s)" % (s[0], s[1], to_id(s[0])) for s in self.uses[f["name"]]])
+        #    doc += ", ".join([" [%s.%s](#%s)" % (s[0], s[1], to_id(s[0]))
+        #       for s in self.uses[f["name"]]])
 
         doc = doc + "\n\n" + f["doc"]
 
@@ -317,16 +356,21 @@ class RenderType(object):
                     opt = True
 
                 desc = i["doc"]
-                #if "inherited_from" in i:
+                # if "inherited_from" in i:
                 #    desc = "%s _Inherited from %s_" % (desc, linkto(i["inherited_from"]))
 
-                frg = schema.avro_name(i["name"])
-                tr = "<td><code>%s</code></td><td>%s</td><td>%s</td><td>%s</td>" % (frg, self.typefmt(tp, self.redirects), opt, mistune.markdown(desc))
+                rfrg = schema.avro_name(i["name"])
+                tr = "<td><code>%s</code></td><td>%s</td><td>%s</td>"\
+                    "<td>%s</td>" % (
+                        rfrg, self.typefmt(tp, self.redirects,
+                                           jsonldPredicate=i.get("jsonldPredicate")),
+                        opt,
+                        mistune.markdown(desc))
                 if opt:
                     required.append(tr)
                 else:
                     optional.append(tr)
-            for i in required+optional:
+            for i in required + optional:
                 doc += "<tr>" + i + "</tr>"
             doc += """</table>"""
         elif f["type"] == "enum":
@@ -336,30 +380,34 @@ class RenderType(object):
             for e in ex:
                 for i in e.get("symbols", []):
                     doc += "<tr>"
-                    frg = schema.avro_name(i)
-                    doc += "<td><code>%s</code></td><td>%s</td>" % (frg, enumDesc.get(frg, ""))
+                    efrg = schema.avro_name(i)
+                    doc += "<td><code>%s</code></td><td>%s</td>" % (
+                        efrg, enumDesc.get(efrg, ""))
                     doc += "</tr>"
             doc += """</table>"""
         f["doc"] = doc
 
         self.typedoc.write(f["doc"])
 
-        subs = self.docParent.get(f["name"], []) + self.record_refs.get(f["name"], [])
+        subs = self.docParent.get(f["name"], []) + \
+            self.record_refs.get(f["name"], [])
         if len(subs) == 1:
             self.render_type(self.typemap[subs[0]], depth)
         else:
             for s in subs:
-                self.render_type(self.typemap[s], depth+1)
+                self.render_type(self.typemap[s], depth + 1)
 
         for s in self.docAfter.get(f["name"], []):
             self.render_type(self.typemap[s], depth)
 
+
 def avrold_doc(j, outdoc, renderlist, redirects, brand, brandlink):
+    # type: (List[Dict[unicode, Any]], IO[Any], str, Dict, str, str) -> None
     toc = ToC()
     toc.start_numbering = False
 
     rt = RenderType(toc, j, renderlist, redirects)
-    content = rt.typedoc.getvalue()
+    content = rt.typedoc.getvalue()  # type: unicode
 
     outdoc.write("""
     <!DOCTYPE html>
@@ -383,6 +431,10 @@ def avrold_doc(j, outdoc, renderlist, redirects, brand, brandlink):
     .tocnav ol {
       list-style: none
     }
+    pre {
+      margin-left: 2em;
+      margin-right: 2em;
+    }
     </style>
     </head>
     <body>
@@ -430,6 +482,7 @@ def avrold_doc(j, outdoc, renderlist, redirects, brand, brandlink):
     </body>
     </html>""")
 
+
 if __name__ == "__main__":
 
     parser = argparse.ArgumentParser()
@@ -442,25 +495,28 @@ if __name__ == "__main__":
 
     args = parser.parse_args()
 
-    s = []
+    s = []  # type: List[Dict[unicode, Any]]
     a = args.schema
     with open(a) as f:
         if a.endswith("md"):
             s.append({"name": os.path.splitext(os.path.basename(a))[0],
-                  "type": "documentation",
-                  "doc": f.read().decode("utf-8")
-              })
+                      "type": "documentation",
+                      "doc": f.read().decode("utf-8")
+                      })
         else:
             uri = "file://" + os.path.abspath(a)
-            _, _, metaschema_loader = schema.get_metaschema()
+            metaschema_loader = schema.get_metaschema()[2]
             j, schema_metadata = metaschema_loader.resolve_ref(uri, "")
             if isinstance(j, list):
                 s.extend(j)
-            else:
+            elif isinstance(j, dict):
                 s.append(j)
+            else:
+                raise ValueError("Schema must resolve to a list or a dict")
 
     primitiveType = args.primtype
-
-    redirect = {r.split("=")[0]:r.split("=")[1] for r in args.redirect} if args.redirect else {}
+    redirect = {}
+    for r in (args.redirect or []):
+        redirect[r.split("=")[0]] = r.split("=")[1]
     renderlist = args.only if args.only else []
     avrold_doc(s, sys.stdout, renderlist, redirect, args.brand, args.brandlink)
